/***************************************************************************

  osdepend.c

  OS dependant stuff (display handling, keyboard scan...)
  This is the only file which should me modified in order to port the
  emulator to a different system.

***************************************************************************/

#include "osdepend.h"
#include "../osdepend.h"
#include "../mame.h"
#include "../driver.h"
//#include "sound.h"
#include <kos.h>
#include <dc/sound/stream.h>
#include <dc/biosfont.h>
#include <time.h>
#include "unzip.h"

/* frameskip stuff */
static int frameskip=0;
static int frameskip_setting=0,time_frameskip=-1,nb_frames=0,frameskip_delay=0;
static int use_auto_frameskip=1, use_sound=0, use_antialias=0;


int	play_sound = 0;
int key_port,key_unit;
kbd_state_t *state;

int gfx_width;
int gfx_height;

cont_cond_t 	ccond[4];

uint8 maple_joystick_addr[4] = {0, 0, 0, 0};

int  raise(int signal)
{
	printf("raise\n");
	return -1 ;
}
#define DC_TEXTURE_ADDRS_SIZE  2
static pvr_ptr_t dc_texture_addrs[DC_TEXTURE_ADDRS_SIZE];
static pvr_poly_hdr_t dc_pvr_poly_headers_filter_none[DC_TEXTURE_ADDRS_SIZE];
static pvr_poly_hdr_t dc_pvr_poly_headers_filter_bilinear[DC_TEXTURE_ADDRS_SIZE];

typedef struct { 
  float u1, u2 ,u3,x1,x2,y1,y2;
  pvr_poly_hdr_t *poly_headers;
} dc_pvr_setting_t; 

static dc_pvr_setting_t dc_pvr_setting[DC_TEXTURE_ADDRS_SIZE];

void
dc_set_pvr_setting ( float u1, float u2, float u3,float x1,float x2, float y1,float y2)
{
  dc_pvr_setting_t *p = dc_pvr_setting ;
  
  if (0 <= u1 && u1 <= 1) p->u1 = u1;
  if (0 <= u2 && u2 <= 1) p->u2 = u2;
  if (0 <= u3 && u3 <= 1) p->u3 = u3;
   p->x1 = x1;
   p->x2 = x2;
   p->y1 = y1;
   p->y2 = y2;
//	if(use_antialias)
   		p->poly_headers = dc_pvr_poly_headers_filter_none;
//	else
//		p->poly_headers = dc_pvr_poly_headers_filter_bilinear;
}

static void
dc_draw_screen(int index)
{
  pvr_vertex_t   vert;
  dc_pvr_setting_t *setting = dc_pvr_setting ;
  
  pvr_wait_ready ();

  pvr_scene_begin ();
  pvr_list_begin (PVR_LIST_OP_POLY);
  
  pvr_prim (setting->poly_headers +index, sizeof(pvr_poly_hdr_t));
  pvr_dr_init(index);

  vert.flags = PVR_CMD_VERTEX;
  vert.x = setting->x1;
  vert.y = setting->y2;
  vert.z = 1.0f;
  vert.u = setting->u1;
  vert.v = setting->u3;
  vert.argb = PVR_PACK_COLOR (1, 1, 1, 1);

  vert.oargb = 0;
  pvr_prim (&vert, sizeof(vert));
  
  vert.x = setting->x1;
  vert.y = setting->y1;
  vert.u = setting->u1;
  vert.v = 0;
  pvr_prim (&vert, sizeof(vert));
  
  vert.x = setting->x2;
  vert.y = setting->y2;
  vert.u = setting->u2;
  vert.v = setting->u3;
  pvr_prim (&vert, sizeof(vert));
  
  vert.flags = PVR_CMD_VERTEX_EOL;
  vert.x = setting->x2;
  vert.y =0;
  vert.u = setting->u2;
  vert.v = 0;
  pvr_prim (&vert, sizeof(vert));
  pvr_dr_commit(index);

  pvr_list_finish ();
  
  pvr_scene_finish ();
}

static pvr_init_params_t pvr_params = 
{
  /* Enable opaque polygons with size 16 */
  { PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0, PVR_BINSIZE_0, PVR_BINSIZE_0 },
  
  /* Vertex buffer size 512K */
  512 *1024
};

void
dc_vid_init ()
{
  int i;
  pvr_poly_cxt_t poly;
  
  pvr_init (&pvr_params); 
  for (i = 0; i < DC_TEXTURE_ADDRS_SIZE; ++i) {
    pvr_ptr_t addr;
    
    addr = pvr_mem_malloc ( gfx_width * gfx_height * 2);
 //  printf("gfx_width * gfx_height = %d * %d\n",gfx_width ,gfx_height );
    dc_texture_addrs [i] = addr; 
    
    pvr_poly_cxt_txr (&poly, PVR_LIST_OP_POLY, 
                      PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED,
                      gfx_width, gfx_height, addr,PVR_FILTER_NONE);
    pvr_poly_compile (dc_pvr_poly_headers_filter_none +i, &poly);
    
  }

  memset (dc_pvr_setting, 0, sizeof(dc_pvr_setting));

}


void display_loading(int now, int max);

static void updatekeyboard(void)
{
	state=kbd_get_state(key_port,key_unit);
}


static int last_check = 0;
void poll_joystick(void)
{
	
	static int count = 0;
	if (last_check == maple_state.vbl_cntr)
		return ;
	
	last_check = maple_state.vbl_cntr;

	updatekeyboard() ;

//	if (count != 2)
	{
		// read the joysticks, one at a time
		if (maple_joystick_addr[count])
		{
//		printf("contgetcond")+
		cont_get_cond(maple_joystick_addr[count], &ccond[count]);
		}
	}
	
	// increment and wrap
	count = (count + 1) & 1;
}

int key_port,key_unit;

int mamedc_init()
{
	uint8 key_device ;

	key_device= maple_first_kb();
	key_port= (key_device >> 6 );
	key_unit= (key_device & 0x1f);

	int port, unit, tot;

	// find as many as four controllers
	tot = 0;
		for (port = 0; port < 4; port++)
		{
			for (unit = 0; unit < 6; unit++) 
			{
				if (maple_device_func(port, unit) & MAPLE_FUNC_CONTROLLER) 
				{
					maple_joystick_addr[tot] = maple_addr(port, unit);
					tot++;
				}
			}
		}
	
	return 0;

}

void mamedc_exit()
{
		snd_stream_shutdown();

}

/* put here anything you need to do when the program is started. Return 0 if */
/* initialization was successful, nonzero otherwise. */
int osd_init(void)
{
  frameskip=0;
  time_frameskip=-1; nb_frames=0;
  frameskip_delay=0;
  printf("osd init called\r\n");
	return 0;
}

/* put here cleanup routines to be executed when the program is terminated. */
void osd_exit(void)
{
}
const int safety = 16;

struct osd_bitmap *osd_alloc_bitmap(int width,int height,int depth)
{
	struct osd_bitmap *bitmap;

  printf("osd_alloc_bitmap called %i %i %i\r\n",width,height,depth);

	if (depth != 8 && depth != 15 && depth != 16 && depth != 32)
	{
		printf("osd_alloc_bitmap() unknown depth %d\n",depth);
		return NULL;
	}

	if ((bitmap = malloc(sizeof(struct osd_bitmap))) != 0)
	{
		int i,rowlen,rdwidth;
		unsigned char *bm;

		bitmap->depth = depth;
		bitmap->width = width;
		bitmap->height = height;

		rdwidth = (width + 7) & ~7;     /* round width to a quadword */
		rowlen = (rdwidth + 2 * safety) * sizeof(unsigned char);
		if (depth == 32)
			rowlen *= 4;
		else if (depth == 15 || depth == 16)
			rowlen *= 2;

		if ((bm = malloc((height + 2 * safety) * rowlen)) == 0)
		{
			free(bitmap);
			return 0;
		}

		/* clear ALL bitmap, including safety area, to avoid garbage on right */
		/* side of screen if width is not a multiple of 4 */
		memset(bm,0,(height + 2 * safety) * rowlen);

		if ((bitmap->line = malloc((height + 2 * safety) * sizeof(unsigned char *))) == 0)
		{
			free(bm);
			free(bitmap);
			return 0;
		}

		for (i = 0;i < height + 2 * safety;i++)
		{
			if (depth == 32)
				bitmap->line[i] = &bm[i * rowlen + 4*safety];
			else if (depth == 15 || depth == 16)
				bitmap->line[i] = &bm[i * rowlen + 2*safety];
			else
				bitmap->line[i] = &bm[i * rowlen + safety];
		}
		bitmap->line += safety;

		bitmap->_private = bm;
	}

	return bitmap;
}

void osd_free_bitmap(struct osd_bitmap *bitmap)
{
	if (bitmap)
	{
		bitmap->line -= safety;
		free(bitmap->line);
		free(bitmap->_private);
		free(bitmap);
	}
}


int osd_create_display(int width,int height,int depth,int fps,int attributes,int orientation)
{
  printf("create display %i %i %i\r\n",width,height,depth);

#define MIN_TA_SIZE 256
			gfx_width = (width / MIN_TA_SIZE) * MIN_TA_SIZE;
			gfx_height = (height / MIN_TA_SIZE) * MIN_TA_SIZE;

			if (width > gfx_width) 
				gfx_width += MIN_TA_SIZE;
			if (height > gfx_height) 
				gfx_height+=MIN_TA_SIZE;

			float x1 = 0;
			float x2 = 640 ;
		        float y1 = 0;
			float y2 = 480 ;	

			dc_vid_init ();
			dc_set_pvr_setting (0 , (float)width/gfx_width, (float)height/gfx_height,x1,x2,y1,y2);


  return 0;
}

void osd_close_display(void)
{
	pvr_shutdown();

}

int vis_min_x,vis_min_y,vis_max_x,vis_max_y;
int vis_width, vis_height;
void osd_set_visible_area(int min_x,int max_x,int min_y,int max_y)
{
	// copy the new parameters
	vis_min_x = min_x;
	vis_max_x = max_x;
	vis_min_y = min_y;
	vis_max_y = max_y;

	// track these changes
	printf("osd_set visible area %d-%d %d-%d\r\n",min_x,max_x,min_y,max_y);

	// compute the visible width and height
	vis_width  = max_x - min_x + 1;
	vis_height = max_y - min_y + 1;

	// tell the UI where it can draw
	set_ui_visarea(vis_min_x, vis_min_y, vis_min_x + vis_width - 1, vis_min_y + vis_height - 1);

	// now adjust the window for the aspect ratio
/*	if (vis_width > 1 && vis_height > 1)
		adjust_window_for_visible(min_x, max_x, min_y, max_y);*/
}

#define MAX_PEN 256
int first_free_pen = 0;

unsigned short pens[MAX_PEN];

int osd_allocate_colors(unsigned int totalcolors,
		const UINT8 *palette,UINT32 *pns,int modifiable,
		const UINT8 *debug_palette,UINT32 *debug_pens)
{
  int i;
  int screen_colors=MAX_PEN;

  printf("osd_allocate_colors called %i\r\n",totalcolors);

	pens[1] = 0xffff;

			Machine->uifont->colortable[0] = 0;
			Machine->uifont->colortable[1] = 1;
			Machine->uifont->colortable[2] = 1;
			Machine->uifont->colortable[3] = 0;

			/* fill the palette starting from the end, so we mess up badly written */
			/* drivers which don't go through Machine->pens[] */
			for (i = 0;i < totalcolors;i++)
				pns[i] = (screen_colors-1)-i;

	for (i = 0;i < totalcolors;i++)
		{
			pens[pns[i]] = ((palette[3*i]>>3)<<11) + ((palette[(3*i)+1]>>2)<<5) + (palette[(3*i)+2]>>3);
		}

  return 0;
}
void osd_modify_pen(int pen,unsigned char red, unsigned char green, unsigned char blue)
{
//  printf("osd_modify_pen called %i %i %i %i\r\n",pen,red,green,blue);
  pens[pen] = ((red >> 3) << 11) | ((green >> 2) << 5) | ((blue >> 3) << 0);
}


void osd_update_video_and_audio(
		struct osd_bitmap *bitmap,struct osd_bitmap *debug_bitmap,int leds_status)
{
	register unsigned short * walk_display;
	register unsigned char * walk_bitmap;
	register unsigned short * walk_bitmap16;

	register int i;
	register int j;

/*  if(use_auto_frameskip)
  {
    if(frameskip_delay<50)
      frameskip_delay++;
    else
    {
      if(time_frameskip!=-1)
      {
        int t=thd->jiffies()-time_frameskip;
        float a,t2;
        nb_frames++;
//    t2=(t*60)/50;
//    a=t2/nb_frames;
        a=t/nb_frames;
//    printf("jif: %i skip: %i ratio: %i\r\n",thd->jiffies(), frameskip_setting, (int)(a*1000));
        if(a>1.2f) frameskip_setting++;
        if(a<1.2f) frameskip_setting--;
        if(frameskip_setting<0) frameskip_setting=0;
        if(frameskip_setting>11) frameskip_setting=11;
      } else time_frameskip=thd->jiffies();
  
      if((frameskip++)!=0)
      {
        if(frameskip>=frameskip_setting) frameskip=0;
        else return;
      }
    }
  }*/

	static int texture_index = 0;
		for (i =0 ;i < bitmap->height;i++)
		{
			walk_bitmap = bitmap->line[i] ;
			walk_display = dc_texture_addrs[texture_index] + i * (gfx_width * 2);
			for (j =0 ; j < bitmap->width; j++)
			{
				*walk_display++ = pens[*walk_bitmap++];
			}
			
		}
 dc_draw_screen(texture_index);
  texture_index = (texture_index + 1) & (DC_TEXTURE_ADDRS_SIZE - 1);

	poll_joystick();

}

void osd_get_pen(int pen,unsigned char *red, unsigned char *green, unsigned char *blue)
{
  printf("osd_get_pen called\r\n");
}

void osd_mark_dirty(int xmin,int ymin,int xmax,int ymax)
{
//  printf("osd_mark_dirty called %i %i %i %i\r\n",xmin,ymin,xmax,ymax);
}

int osd_skip_this_frame(void)
{
  return frameskip!=0;
}


void osd_debugger_focus(int debugger_has_focus)
{
}

void osd_set_gamma(float _gamma)
{
}

float osd_get_gamma(void)
{
  printf("osd_get_gamma\r\n");
  return 1.0;
}

void osd_set_brightness(int brightness)
{
}

int osd_get_brightness(void)
{
  printf("osd_get_brightness\r\n");
  return 100;
}

void osd_save_snapshot(struct osd_bitmap *bitmap)
{
}

// keyboard/controller stuff

#define KBD_KEY_LSHIFT 0x70
#define KBD_KEY_RSHIFT 0x71
#define KBD_KEY_LCONTROL 0x72
#define KBD_KEY_RCONTROL 0x73
#define KBD_KEY_ALT 0x74
#define KBD_KEY_ALTGR 0x75

static struct KeyboardInfo keylist[] =
{
	{ "A",			KBD_KEY_A,				KEYCODE_A },
	{ "B",			KBD_KEY_B,				KEYCODE_B },
	{ "C",			KBD_KEY_C,				KEYCODE_C },
	{ "D",			KBD_KEY_D,				KEYCODE_D },
	{ "E",			KBD_KEY_E,				KEYCODE_E },
	{ "F",			KBD_KEY_F,				KEYCODE_F },
	{ "G",			KBD_KEY_G,				KEYCODE_G },
	{ "H",			KBD_KEY_H,				KEYCODE_H },
	{ "I",			KBD_KEY_I,				KEYCODE_I },
	{ "J",			KBD_KEY_J,				KEYCODE_J },
	{ "K",			KBD_KEY_K,				KEYCODE_K },
	{ "L",			KBD_KEY_L,				KEYCODE_L },
	{ "M",			KBD_KEY_M,				KEYCODE_M },
	{ "N",			KBD_KEY_N,				KEYCODE_N },
	{ "O",			KBD_KEY_O,				KEYCODE_O },
	{ "P",			KBD_KEY_P,				KEYCODE_P },
	{ "Q",			KBD_KEY_Q,				KEYCODE_Q },
	{ "R",			KBD_KEY_R,				KEYCODE_R },
	{ "S",			KBD_KEY_S,				KEYCODE_S },
	{ "T",			KBD_KEY_T,				KEYCODE_T },
	{ "U",			KBD_KEY_U,				KEYCODE_U },
	{ "V",			KBD_KEY_V,				KEYCODE_V },
	{ "W",			KBD_KEY_W,				KEYCODE_W },
	{ "X",			KBD_KEY_X,				KEYCODE_X },
	{ "Y",			KBD_KEY_Y,				KEYCODE_Y },
	{ "Z",			KBD_KEY_Z,				KEYCODE_Z },
	{ "0",			KBD_KEY_0,				KEYCODE_0 },
	{ "1",			KBD_KEY_1,				KEYCODE_1 },
	{ "2",			KBD_KEY_2,				KEYCODE_2 },
	{ "3",			KBD_KEY_3,				KEYCODE_3 },
	{ "4",			KBD_KEY_4,				KEYCODE_4 },
	{ "5",			KBD_KEY_5,				KEYCODE_5 },
	{ "6",			KBD_KEY_6,				KEYCODE_6 },
	{ "7",			KBD_KEY_7,				KEYCODE_7 },
	{ "8",			KBD_KEY_8,				KEYCODE_8 },
	{ "9",			KBD_KEY_9,				KEYCODE_9 },
	{ "0 PAD",		KBD_KEY_PAD_0,			KEYCODE_0_PAD },
	{ "1 PAD",		KBD_KEY_PAD_1,			KEYCODE_1_PAD },
	{ "2 PAD",		KBD_KEY_PAD_2,			KEYCODE_2_PAD },
	{ "3 PAD",		KBD_KEY_PAD_3,			KEYCODE_3_PAD },
	{ "4 PAD",		KBD_KEY_PAD_4,			KEYCODE_4_PAD },
	{ "5 PAD",		KBD_KEY_PAD_5,			KEYCODE_5_PAD },
	{ "6 PAD",		KBD_KEY_PAD_6,			KEYCODE_6_PAD },
	{ "7 PAD",		KBD_KEY_PAD_7,			KEYCODE_7_PAD },
	{ "8 PAD",		KBD_KEY_PAD_8,			KEYCODE_8_PAD },
	{ "9 PAD",		KBD_KEY_PAD_9,			KEYCODE_9_PAD },
	{ "F1",			KBD_KEY_F1,				KEYCODE_F1 },
	{ "F2",			KBD_KEY_F2,				KEYCODE_F2 },
	{ "F3",			KBD_KEY_F3,				KEYCODE_F3 },
	{ "F4",			KBD_KEY_F4,				KEYCODE_F4 },
	{ "F5",			KBD_KEY_F5,				KEYCODE_F5 },
	{ "F6",			KBD_KEY_F6,				KEYCODE_F6 },
	{ "F7",			KBD_KEY_F7,				KEYCODE_F7 },
	{ "F8",			KBD_KEY_F8,				KEYCODE_F8 },
	{ "F9",			KBD_KEY_F9,				KEYCODE_F9 },
	{ "F10",		KBD_KEY_F10,			KEYCODE_F10 },
	{ "F11",		KBD_KEY_F11,			KEYCODE_F11 },
	{ "F12",		KBD_KEY_F12,			KEYCODE_F12 },
	{ "ESC",		KBD_KEY_ESCAPE,			KEYCODE_ESC },
	{ "~",			KBD_KEY_TILDE,			KEYCODE_TILDE },
	{ "-",          KBD_KEY_MINUS,          KEYCODE_MINUS },
//	{ "=",          KBD_KEY_EQUALS,         KEYCODE_EQUALS },
	{ "BKSPACE",	KBD_KEY_BACKSPACE,		KEYCODE_BACKSPACE },
	{ "TAB",		KBD_KEY_TAB,			KEYCODE_TAB },
	{ "[",          KBD_KEY_LBRACKET,      KEYCODE_OPENBRACE },
	{ "]",          KBD_KEY_RBRACKET,     KEYCODE_CLOSEBRACE },
	{ "ENTER",		KBD_KEY_ENTER,			KEYCODE_ENTER },
	{ ";",          KBD_KEY_SEMICOLON,          KEYCODE_COLON },
	{ ":",          KBD_KEY_QUOTE,          KEYCODE_QUOTE },
	{ "\\",         KBD_KEY_BACKSLASH,      KEYCODE_BACKSLASH },
//	{ "<",          KBD_KEY_BACKSLASH2,     KEYCODE_BACKSLASH2 },
	{ ",",          KBD_KEY_COMMA,          KEYCODE_COMMA },
	{ ".",          KBD_KEY_PERIOD,           KEYCODE_STOP },
	{ "/",          KBD_KEY_SLASH,          KEYCODE_SLASH },
	{ "SPACE",		KBD_KEY_SPACE,			KEYCODE_SPACE },
	{ "INS",		KBD_KEY_INSERT,			KEYCODE_INSERT },
	{ "DEL",		KBD_KEY_DEL,			KEYCODE_DEL },
	{ "HOME",		KBD_KEY_HOME,			KEYCODE_HOME },
	{ "END",		KBD_KEY_END,			KEYCODE_END },
	{ "PGUP",		KBD_KEY_PGUP,			KEYCODE_PGUP },
	{ "PGDN",		KBD_KEY_PGDOWN,			KEYCODE_PGDN },
	{ "LEFT",		KBD_KEY_LEFT,			KEYCODE_LEFT },
	{ "RIGHT",		KBD_KEY_RIGHT,			KEYCODE_RIGHT },
	{ "UP",			KBD_KEY_UP,				KEYCODE_UP },
	{ "DOWN",		KBD_KEY_DOWN,			KEYCODE_DOWN },
	{ "/ PAD",      KBD_KEY_PAD_DIVIDE,      KEYCODE_SLASH_PAD },
	{ "* PAD",      KBD_KEY_PAD_MULTIPLY,       KEYCODE_ASTERISK },
	{ "- PAD",      KBD_KEY_PAD_MINUS,      KEYCODE_MINUS_PAD },
	{ "+ PAD",      KBD_KEY_PAD_PLUS,       KEYCODE_PLUS_PAD },
	{ ". PAD",      KBD_KEY_PAD_PERIOD,        KEYCODE_DEL_PAD },
	{ "ENTER PAD",  KBD_KEY_PAD_ENTER,      KEYCODE_ENTER_PAD },
	{ "PRTSCR",     KBD_KEY_PRINT,         KEYCODE_PRTSCR },
	{ "PAUSE",      KBD_KEY_PAUSE,          KEYCODE_PAUSE },
	{ "LSHIFT",		KBD_KEY_LSHIFT,			KEYCODE_LSHIFT },
	{ "RSHIFT",		KBD_KEY_RSHIFT,			KEYCODE_RSHIFT },
	{ "LCTRL",		KBD_KEY_LCONTROL,		KEYCODE_LCONTROL },
	{ "RCTRL",		KBD_KEY_RCONTROL,		KEYCODE_RCONTROL },
	{ "ALT",		KBD_KEY_ALT,			KEYCODE_LALT },
	{ "ALTGR",		KBD_KEY_ALTGR,			KEYCODE_RALT },
//	{ "LWIN",		KBD_KEY_LWIN,			KEYCODE_OTHER },
//	{ "RWIN",		KBD_KEY_RWIN,			KEYCODE_OTHER },
//	{ "MENU",		KBD_KEY_MENU,			KEYCODE_OTHER },
	{ "SCRLOCK",    KBD_KEY_SCRLOCK,        KEYCODE_SCRLOCK },
	{ "NUMLOCK",    KBD_KEY_PAD_NUMLOCK,        KEYCODE_NUMLOCK },
	{ "CAPSLOCK",   KBD_KEY_CAPSLOCK,       KEYCODE_CAPSLOCK },
	{ 0, 0, 0 }	/* end of table */
};

const struct KeyboardInfo *osd_get_key_list(void)
{
  return keylist;
}

/*
  tell whether the specified key is pressed or not. keycode is the OS dependant
  code specified in the list returned by osd_get_key_list().
*/
int osd_is_key_pressed(int keycode)
{
  int i;

	if ((keycode == KBD_KEY_LSHIFT) && (state->cond.modifiers & KBD_MOD_LSHIFT))
		return 1;
	if ((keycode == KBD_KEY_RSHIFT) && (state->cond.modifiers & KBD_MOD_RSHIFT))
		return 1;
	if ((keycode == KBD_KEY_LCONTROL) && (state->cond.modifiers & KBD_MOD_LCTRL))
		return 1;
	if ((keycode == KBD_KEY_RCONTROL) && (state->cond.modifiers & KBD_MOD_RCTRL))
		return 1;
	if ((keycode == KBD_KEY_ALT) && (state->cond.modifiers & KBD_MOD_LALT))
		return 1;
	if ((keycode == KBD_KEY_ALTGR) && (state->cond.modifiers & KBD_MOD_RALT))
		return 1;

		for (i = 0; i < 6; i++)
		{
			if (state->cond.keys[i] == keycode)
			{
				return 1;
			}
		}
	// add all joystick keys here
	if (maple_joystick_addr[0])
	{
	// special case for ESC
	if((keycode==KBD_KEY_ESCAPE) && !(ccond[0].buttons & CONT_A) 
		 && !(ccond[0].buttons & CONT_START))
	{
		// wait for release before returning
		do
		{
			poll_joystick();
			timer_spin_sleep(100);
		}
		while (!(ccond[0].buttons & CONT_A) 
			 || !(ccond[0].buttons & CONT_START));
		
		return 1;
	}

		// coins
		if ((keycode==KBD_KEY_5) && (ccond[0].rtrig > 16))
			return 1;
		
		// player
		if ((keycode==KBD_KEY_1) && !(ccond[0].buttons & CONT_START))
			return 1;
		
		// up
		if ((keycode==KBD_KEY_UP) && !(ccond[0].buttons & CONT_DPAD_UP))
			return 1;
	//	if ((keycode==KBD_KEY_UP) && ccond[0].joyy<(128-16))
	//		return 1;
		
		// down
		if ((keycode==KBD_KEY_DOWN) && !(ccond[0].buttons & CONT_DPAD_DOWN))
			return 1;
	//	if ((keycode==KBD_KEY_DOWN) && ccond[0].joyy>(128+16))
	//		return 1;
		
		// left
		if ((keycode==KBD_KEY_LEFT) && !(ccond[0].buttons & CONT_DPAD_LEFT))
			return 1;
	//	if ((keycode==KBD_KEY_LEFT) && ccond[0].joyx<(128-16))
	//		return 1;
		
		// right	
		if ((keycode==KBD_KEY_RIGHT) && !(ccond[0].buttons & CONT_DPAD_RIGHT))
			return 1;
	//	if ((keycode==KBD_KEY_RIGHT) && ccond[0].joyx>(128+16))
	//		return 1;
			
		// button 1	
		if ((keycode==KBD_KEY_LCONTROL) && !(ccond[0].buttons & CONT_A))
			return 1;

		// button 2
		if ((keycode==KBD_KEY_ALT) && !(ccond[0].buttons & CONT_X))
			return 1;
			
		// button 3
		if ((keycode==KBD_KEY_SPACE) && !(ccond[0].buttons & CONT_B))
			return 1;
		
		// button 4
		if ((keycode==KBD_KEY_LSHIFT) && !(ccond[0].buttons & CONT_Y))
			return 1;

		// UI toggle
		if ((keycode==KBD_KEY_TAB) && (ccond[0].ltrig > 16))
			return 1;

		// UI select
		if ((keycode==KBD_KEY_ENTER) && !(ccond[0].buttons & CONT_A))
			return 1;
	}

	if (maple_joystick_addr[1])
	{
		if ((keycode==KBD_KEY_6) && (ccond[1].rtrig > 16))
			return 1;

		if ((keycode==KBD_KEY_2) && !(ccond[1].buttons & CONT_START))
			return 1;

		if ((keycode==KBD_KEY_R) && !(ccond[1].buttons & CONT_DPAD_UP))
			return 1;
	//	if ((keycode==KBD_KEY_R) && ccond[1].joyy<(128-16))
	//		return 1;

		if ((keycode==KBD_KEY_F) && !(ccond[1].buttons & CONT_DPAD_DOWN))
			return 1;
	//	if ((keycode==KBD_KEY_F) && ccond[1].joyy>(128+16))
	//		return 1;

		if ((keycode==KBD_KEY_D) && !(ccond[1].buttons & CONT_DPAD_LEFT))
			return 1;
	//	if ((keycode==KBD_KEY_D) && ccond[1].joyx<(128-16))
	//		return 1;

		if ((keycode==KBD_KEY_G) && !(ccond[1].buttons & CONT_DPAD_RIGHT))
			return 1;
	//	if ((keycode==KBD_KEY_G) && ccond[1].joyx>(128+16))
	//		return 1;

		if ((keycode==KBD_KEY_A) && !(ccond[1].buttons & CONT_A))
			return 1;

		if ((keycode==KBD_KEY_S) && !(ccond[1].buttons & CONT_X))
			return 1;

		if ((keycode==KBD_KEY_Q) && !(ccond[1].buttons & CONT_B))
			return 1;

		if ((keycode==KBD_KEY_W) && !(ccond[1].buttons & CONT_Y))
			return 1;
	}

  return 0;
}

/*
  Return the Unicode value of the most recently pressed key. This
  function is used only by text-entry routines in the user interface and should
  not be used by drivers. The value returned is in the range of the first 256
  bytes of Unicode, e.g. ISO-8859-1. A return value of 0 indicates no key down.

  Set flush to 1 to clear the buffer before entering text. This will avoid
  having prior UI and game keys leak into the text entry.
*/
int osd_readkey_unicode(int flush)
{
  return 0;
}

void osd_customize_inputport_defaults(struct ipd *defaults)
{
}

uclock_t uclock(void)
{
  static uclock_t init_sec = 0;
  struct timeval tv;
  
  gettimeofday(&tv, 0);
  if (init_sec == 0) init_sec = tv.tv_sec;
  
  return (tv.tv_sec - init_sec) * 1000000 + tv.tv_usec;
}

// file

int osd_faccess(const char *filename, int filetype)
{
  printf("osd_facces called\r\n");
  return 0;
}

typedef enum
{
	kPlainFile,
	kZippedFile
}	eFileType;

typedef struct
{
	FILE *file;
	unsigned char *data;
	unsigned int offset;
	unsigned int length;
	eFileType type;
	unsigned int crc;
}	FakeFileHandle;

extern unsigned int crc32 (unsigned int crc, const unsigned char *buf, unsigned int len);

void *osd_fopen(const char *gamename,const char *filename,int filetype,int read_or_write)
{
  FakeFileHandle *f;
  f = (FakeFileHandle *) malloc (sizeof (FakeFileHandle));
	if( !f )
	{
		printf("osd_fopen: failed to malloc FakeFileHandle!\n");
    return 0;
	}
	memset (f, 0, sizeof (FakeFileHandle));

  printf("osd_fopen_called: filetype=%i game=%s\r\n",filetype,gamename);
  if(filetype==OSD_FILETYPE_ROM)
  {
    char buf[1024];
    sprintf(buf,"/cd/roms/%s/%s",gamename,filename);

    if(read_or_write)
    {
//      return (void *)fopen(buf,"wb");
    } else {
      void *fh=fopen(buf,"rb");
      if(fh)
      {
        f->type=kPlainFile;
        f->file=fh;
        return f;
      }
      if(gamename[0])
      {
        // try in a zip file
        sprintf(buf,"/cd/roms/%s.zip",gamename);


        if(load_zipped_file(buf,filename,&f->data,&f->length)==0)
        {
//          printf("found in zipped file!\r\n");
          f->type=kZippedFile;
          f->crc=crc32 (0L, f->data, f->length);
          return f;
        }
      }
    }
  }
  free(f);
  return 0;
}

int osd_fread(void *file,void *buffer,int length)
{
  FakeFileHandle *f=(FakeFileHandle *)file;
  if(f->type==kPlainFile) return fread(buffer, length, 1, f->file);
  if( f->data )
	{
	  if( length + f->offset > f->length )
		  length = f->length - f->offset;
		memcpy (buffer, f->offset + f->data, length);
		f->offset += length;
		return length;
	}
  return 0;
}

int osd_fwrite(void *file,const void *buffer,int length)
{
//  return fwrite((void *)buffer, length, 1, file);
  return 0;
}

void osd_fclose(void *file)
{
  FakeFileHandle *f=(FakeFileHandle *)file;
  if(f->type==kPlainFile) fclose(f->file);
  else free(f->data);
  free(f);
}

int osd_fseek(void *file,int offset,int whence)
{
  FakeFileHandle *f=(FakeFileHandle *)file;
  if(f->type==kPlainFile) return fseek(f->file, offset, whence);
  switch( whence )
		{
		case SEEK_SET:
			f->offset = offset;
			break;
		case SEEK_CUR:
			f->offset += offset;
			break;
		case SEEK_END:
			f->offset = f->length + offset;
			break;
		}
  return 0;
}

int osd_fsize(void *file)
{
  FakeFileHandle *f=(FakeFileHandle *)file;
  if(f->type==kPlainFile) 
  {
    int l;
    int pos=ftell(f->file);
    fseek(f->file,0,SEEK_END);
    l=ftell(f->file);
    fseek(f->file,pos,SEEK_SET);
    return l;
  }
  return f->length;
}

unsigned int osd_fcrc(void *file)
{
  FakeFileHandle *f=(FakeFileHandle *)file;
  if(f->type==kPlainFile) return 0;
  return f->crc;
}

int osd_fgetc(void *file)
{
  printf("fgetc called\r\n");
  return 0;
}

int osd_ungetc(int c, void *file)
{
  printf("ungetc called\r\n");
  return 0;
}

char *osd_fgets(char *s, int n, void *file)
{
  printf("fgets called\r\n");
  return 0;
}

int osd_feof(void *file)
{
  printf("feof called\r\n");
  return 0;
//  return feof(file);
}

int osd_ftell(void *file)
{
  FakeFileHandle *f=(FakeFileHandle *)file;
  if(f->type==kPlainFile) return ftell(f->file);
  return -1L;
}

int osd_fchecksum(const char *gamename, const char *filename, unsigned int *length, unsigned int *sum)
{
  return 0;
}

int osd_display_loading_rom_message(const char *name,int current,int total)
{
  printf("loading rom: %i/%i",current,total);
  display_loading(current,total);
  return 0;
}

void osd_pause(int paused)
{
}

int osd_fread_swap (void *file, void *buffer, int length)
{
	int i;
	unsigned char *buf;
	unsigned char temp;
	int res;


	res = osd_fread (file, buffer, length);

	buf = buffer;
	for( i = 0; i < length; i += 2 )
	{
		temp = buf[i];
		buf[i] = buf[i + 1];
		buf[i + 1] = temp;
	}

	return res;
}

int osd_fwrite_swap (void *file, const void *buffer, int length)
{
	int i;
	unsigned char *buf;
	unsigned char temp;
	int res;


	buf = (unsigned char *) buffer;
	for( i = 0; i < length; i += 2 )
	{
		temp = buf[i];
		buf[i] = buf[i + 1];
		buf[i + 1] = temp;
	}

	res = osd_fwrite (file, buffer, length);

	for( i = 0; i < length; i += 2 )
	{
		temp = buf[i];
		buf[i] = buf[i + 1];
		buf[i + 1] = temp;
	}

	return res;
}



void CLIB_DECL logerror(const char *text,...)
{
#if 0
	va_list arg;
	va_start(arg,text);
  vprintf(text,arg);
	va_end(arg);
#endif
}

/*
  return a list of all available joystick inputs (see input.h)
*/
const struct JoystickInfo *osd_get_joy_list(void)
{
  return 0;
}

/*
  tell whether the specified joystick direction/button is pressed or not.
  joycode is the OS dependant code specified in the list returned by
  osd_get_joy_list().
*/
int osd_is_joy_pressed(int joycode)
{
  return 0;
}


/* Joystick calibration routines BW 19981216 */
/* Do we need to calibrate the joystick at all? */
int osd_joystick_needs_calibration (void)
{
  return 0;
}

/* Preprocessing for joystick calibration. Returns 0 on success */
void osd_joystick_start_calibration (void)
{
}

/* Prepare the next calibration step. Return a description of this step. */
/* (e.g. "move to upper left") */
const char *osd_joystick_calibrate_next (void)
{
  return "";
}
/* Get the actual joystick calibration data for the current position */
void osd_joystick_calibrate (void)
{
}
/* Postprocessing (e.g. saving joystick data to config) */
void osd_joystick_end_calibration (void)
{
}

void osd_trak_read(int player,int *deltax,int *deltay)
{
}

/* return values in the range -128 .. 128 (yes, 128, not 127) */
void osd_analogjoy_read(int player,int *analog_x, int *analog_y)
{
}

void osd_mark_vector_dirty(int x, int y)
{
}



int fuzzycmp (const char *s, const char *l)
{
	int gaps = 0;
	int match = 0;
	int last = 1;

	for (; *s && *l; l++)
	{
		if (*s == *l)
			match = 1;
		else if (*s >= 'a' && *s <= 'z' && (*s - 'a') == (*l - 'A'))
			match = 1;
		else if (*s >= 'A' && *s <= 'Z' && (*s - 'A') == (*l - 'a'))
			match = 1;
		else
			match = 0;

		if (match)
			s++;

		if (match != last)
		{
			last = match;
			if (!match)
				gaps++;
		}
	}

	/* penalty if short string does not completely fit in */
	for (; *s; s++)
		gaps++;

	return gaps;
}

/* font size constants */
#define FONT_WIDTH 	12
#define FONT_HEIGHT 	24

/* screen size constants */
#define SCREEN_WIDTH	640
#define SCREEN_HEIGHT	480

/* more constants */
#define SCREEN_FIRSTCOL ( (SCREEN_WIDTH - FONT_WIDTH * NUM_COLS) / 2 )
#define SCREEN_LASTCOL  (SCREEN_WIDTH - SCREEN_FIRSTCOL)

/* the font texture */

/* initialize draw stuff: get our texture of the font, etc */
void draw_init() {
	vid_clear(0, 0, 0);
}

int selected=0;
void display_loading(int now, int max)
{
  int i;
  for(i=0;i<10;i++)
  {
      char tmp[128];
      sprintf(tmp,"Loading %s (%i/%i) ...",drivers[selected]->description, now, max);
      bfont_draw_str(vram_s + i * FONT_HEIGHT * 640 , 640, 1, tmp);
  }
}
extern void* kos_audio_callback(int size);

int main (int argc, char **argv)
{
	int i;
  int oldbutt=-1,oldjoy=-1,nb_drv=0,scroll=0,sel=0;
  cont_cond_t ccond;

  	memset(&options, 0, sizeof(options));
	// Init default options
	options.norotate     = 0;
	options.ror          = 0;
	options.rol          = 0;
	options.flipx        = 0;
	options.flipy        = 0;
	options.use_artwork  = 0;
	options.color_depth  = 8;
	options.use_emulated_ym3812 = 1;
	
	options.use_samples  = 1;

	play_sound= 1; 

  if(mamedc_init()) return 0;

  	draw_init();

	snd_stream_init(kos_audio_callback);

  for(i=0;drivers[i];i++) nb_drv++;

  while(1)
  {
	  char tmp[128];


    for(i=0;drivers[i+scroll] && i<11 ;i++)
    {
	sprintf(tmp,"%s%s"," ",(char *)drivers[i+scroll]->description);
	    
      bfont_draw_str(vram_s + i * FONT_HEIGHT * 640 , 640, 1,tmp);
      if((i+scroll)==sel)
      {
	sprintf(tmp,"%s%s","-",(char *)drivers[i+scroll]->description);

      	 bfont_draw_str(vram_s + i * FONT_HEIGHT * 640 , 640, 1, tmp);
	      
      }
    }

    if(drivers[i+scroll]) bfont_draw_str(vram_s + 14 * FONT_HEIGHT * 640 , 640, 1,"!");

    {
      sprintf(tmp,"Frameskip (X): %i  Sound (Y): %s",frameskip_setting, use_sound?"Yes":"No ");
      bfont_draw_str(vram_s + 15 * FONT_HEIGHT * 640 , 640, 1, tmp);
      sprintf(tmp,"Anti-Alias (B): %s",use_antialias?"Yes":"No");
      bfont_draw_str(vram_s + 16 * FONT_HEIGHT * 640 , 640, 1, tmp);
    }


	cont_get_cond(maple_first_controller(), &ccond); 

    if(ccond.buttons!=oldbutt || ccond.joyy!=oldjoy)
    {
      if(!(ccond.buttons & CONT_DPAD_UP))  sel--;
      if(!(ccond.buttons & CONT_DPAD_DOWN)) sel++;
      if(sel<0) sel=0;
      if(sel>=nb_drv) sel=nb_drv-1;
      if(sel>(scroll+10)) scroll++;
      if(sel<scroll) scroll--;
  
      if(!(ccond.buttons & CONT_Y)) use_sound=!use_sound;
  
      if(!(ccond.buttons & CONT_A))
      {
  
      	options.samplerate = use_sound?44100:0;

        run_game(sel);

      }
      oldbutt=ccond.buttons;
      oldjoy=ccond.joyy;
      vid_clear(0, 0, 0);
    }
	if ( (ccond.ltrig > 16) && (ccond.rtrig > 16))
	break;

  }

  mamedc_exit();

  return 0;
}
