#include "driver.h"
#include "dc/sound/stream.h"

//#define AUDIO_BUFFER_SIZE (65536/4)
#define AUDIO_BUFFER_SIZE 10000
#define SMOOTH_SIZE 200

int attenuation = 0;
unsigned char *audio_buffer = NULL;
unsigned char *audio_temp = NULL;

static int audio_stereo;
static int audio_pos = 0; // the position of the next buffer in the audio_buffer to be returned by kcallback

/* global sample tracking */
static int samples_per_frame;
static int samples_per_frame_remainder;
static int samples_left_over;
static int samples_this_frame;

// flag to set for callback to return NULL when stopped
static int stopped = 0;

extern int play_sound ;
// single linear transition from src to dst values
void audio_smooth_together(INT16 *audio_to_smooth, INT16 *audio_to_match)
{
 INT16 * dst_left;
 INT16 * dst_right;
 INT16 * src_left;
 INT16 * src_right;
int i;

 dst_left = audio_to_smooth;
 dst_right = audio_to_smooth + 1;

 src_left = audio_to_match;
 src_right = audio_to_match + 1;

 for (i = 0; i < SMOOTH_SIZE; i++)
 {
  *dst_left =  (((*src_left * (SMOOTH_SIZE - i)) / SMOOTH_SIZE) + ((*dst_left * (i)) / SMOOTH_SIZE));
  *dst_right =  (((*src_right * (SMOOTH_SIZE - i)) / SMOOTH_SIZE) + ((*dst_right * (i)) / SMOOTH_SIZE));

  dst_left += 2;
  dst_right += 2;
  src_left += 2;
  src_right += 2;
 }
}

// simple linear transition from value to dst
void audio_smooth_to_value(INT16 *audio_to_smooth, INT16 value_to_match)
{
 INT16 * dst_left;
 INT16 * dst_right;
int i;

 dst_left = audio_to_smooth;
 dst_right = audio_to_smooth + 1;

 for (i = 0; i < SMOOTH_SIZE; i++)
 {
  *dst_left =  (((value_to_match * (SMOOTH_SIZE - i)) / SMOOTH_SIZE) + ((*dst_left * (i)) / SMOOTH_SIZE));
  *dst_right =  (((value_to_match * (SMOOTH_SIZE - i)) / SMOOTH_SIZE) + ((*dst_right * (i)) / SMOOTH_SIZE));

  dst_left += 2;
  dst_right += 2;
 }
}

// way to much memory moving around for my taste but...
void* kos_audio_callback(int size)
{
 // return NULL if stopped this will cause the sndstream
 // code to silience the sound hardware on the next poll.
 if (stopped)
 {
  return NULL;
 }

// Jim's magic fairy dust is located here
// we don't have enough new data yet so let's manufature some data
// this is where a smaller buffer would be nice to make matching smoother
// but a smaller buffer means that this will occur more frequently
// so it really needs to do a good job
// even at full speed emulation things can get out of sync and we will
// need to manufacture data sometimes
if (size > audio_pos)
{
 // if it's not too bad let's do something fancy
 // otherwise just return the whole last buffer again cause it
 // will match what we will be playing next
 if (size > 2 * audio_pos + SMOOTH_SIZE)
 {
  // well, we are way behind so just repeat the whole last buffer

  // smooth the BACK to match the front here and avoid the seam
  audio_smooth_to_value(audio_temp, audio_temp[size]);

  return audio_temp;
 }
 else
 {
  // copy what we do have into the end of the buffer
  // cause it will match what will will be playing next
  memcpy(audio_temp + (size - audio_pos), audio_buffer, audio_pos);

  // copy what we do have into the beginning of the buffer
  // cause it matches what we have already played
  // but only enough to meet what we copied above
  // leaving only one small seam
  memcpy(audio_temp, audio_buffer, (size - audio_pos));

  // smooth the FRONT to match the back here and avoid the seam
  audio_smooth_together(audio_temp + (size - audio_pos), audio_buffer + (size - audio_pos));

  // we used up all our data so adjust the position
  // hopefully we will have enough next time
  audio_pos = 0;

  return audio_temp;
 }
}

 // copy current audio data into temp buffer to pass back
 memcpy(audio_temp, audio_buffer, size);

 // adjust the position in the audio buffer
 audio_pos -= size;

 // if it didn't ask for all the data slide back the extra
 // data we have to the front of the audio buffer so
 // we know where it is next time through
 if (audio_pos)
 {
  memcpy(audio_buffer, audio_buffer + size, audio_pos);
 }

 // return the new audio data
 return audio_temp;
}

int osd_start_audio_stream(int stereo)
{
	if (Machine->sample_rate == 0) return 0;
//	printf("start\n");
	// first time through allocate the buffers
	if (audio_buffer == 0)
	{
		audio_buffer = (char *)malloc(AUDIO_BUFFER_SIZE * 2);
		audio_temp = (char *)malloc(AUDIO_BUFFER_SIZE);
	}

	memset(audio_buffer, 0, AUDIO_BUFFER_SIZE * 2);
	memset(audio_temp, 0, AUDIO_BUFFER_SIZE);

	// initial position in the audio buffer
	audio_pos  = 0;
	audio_stereo = stereo;

	/* determine the number of samples per frame */
	samples_per_frame = Machine->sample_rate / Machine->drv->frames_per_second;
	samples_per_frame_remainder = Machine->sample_rate % (int)Machine->drv->frames_per_second;
	samples_left_over = samples_per_frame_remainder;
	samples_this_frame = samples_per_frame;

//	printf("samples_per_frame: %d\n", samples_per_frame);
	
	// not stopped anymore
	stopped = 0;

	if (play_sound)
	{
		// start the audio stream, always stereo
		snd_stream_start(options.samplerate, 1);
	}

//	printf("osd_start_audio_stream %d\n", samples_this_frame);
	
	// ask for this many samples for the first frame
	return samples_this_frame;
}

void osd_stop_audio_stream(void)
{
	if (Machine->sample_rate == 0) return;
	
//	printf("osd_stop_audio_stream\n");

	if (audio_buffer != 0) 
	{
		// extra space for main buffer in case we get more than a buffer
		memset(audio_buffer, 0, AUDIO_BUFFER_SIZE * 2);
		// buffer that gets transfered to sndstream
		memset(audio_temp, 0, AUDIO_BUFFER_SIZE);
	}

	stopped = 1;
	
	if (play_sound)
	{
		// wait for silience to get transfered to both buffers before we shutdown
		snd_stream_poll();
		
		// stop the stream
		snd_stream_stop();
	}
	
	free(audio_buffer);
	audio_buffer = 0;
	free(audio_temp);
	audio_temp = 0;
}

int osd_update_audio_stream(INT16 *buffer)
{
	if (Machine->sample_rate == 0) return 0;

	if (play_sound)
	{
		if (audio_stereo)
		{
			// copy the new data into the audio buffer at the current position
			memcpy(audio_buffer + audio_pos, buffer, samples_this_frame * 2);
			audio_pos += samples_this_frame * 2;
		} 
		else 
		{
			int i, j;
			INT16 *p;

			// OK now we have to manufacture stereo data from mono
			i = j = audio_pos / 2;

			p = (INT16 *)audio_buffer;

			for (i = 0; i < samples_this_frame; i++)
			{
				// every other byte
				p[j + (i * 2)]     = *buffer;
				p[j + (i * 2) + 1] = *buffer++;
			}
			
			// and now we have twice the data
			audio_pos += samples_this_frame * 4;
		}
		
		// we have enough data in the buffer so wait for it to be transfered 
		// via the kcallback() before asking for more
	//	if (audio_pos >= AUDIO_BUFFER_SIZE) 
		{
			snd_stream_poll();
		}
	}
	
	// update samples to request for next frame
	samples_left_over = samples_left_over + samples_per_frame_remainder;
	samples_this_frame = samples_per_frame +  samples_left_over / (int)Machine->drv->frames_per_second;
	samples_left_over =  samples_left_over % (int)Machine->drv->frames_per_second;
	
//	printf("osd_start_audio_stream %d\n", samples_this_frame);
	return samples_this_frame;
}

/* attenuation in dB */
void osd_set_mastervolume(int _attenuation)
{
	if (_attenuation > 0) _attenuation = 0;
	if (_attenuation < -32) _attenuation = -32;

//	printf("osd_set_mastervolume %d\n", _attenuation);
	attenuation = _attenuation;
}

int osd_get_mastervolume(void)
{
//	printf("osd_get_mastervolume %d\n", attenuation);
	return attenuation;
}

void osd_sound_enable(int enable_it)
{
}

void osd_opl_control(int chip,int reg)
{
}

void osd_opl_write(int chip,int data)
{
}
